# Run this with ndx-pose 0.1.*
# In ndx-pose 0.2.0, the "nodes" and "edges" constructor arguments of PoseEstimation were deprecated.
# A warning should be raised if they are used when not reading them from a file.
# In ndx-pose 0.2.0, PoseEstimation was changed to raise a DeprecationWarning if the number of original videos,
# labeled videos, or dimensions does not equal the number of camera devices when creating the object not from a file.

import datetime
from ndx_pose import PoseEstimationSeries, PoseEstimation
import numpy as np
from pathlib import Path
from pynwb import NWBFile, NWBHDF5IO


def generate_pose_estimation_nodes_edges():
    """Generate an NWB file with a PoseEstimation object with deprecated fields "nodes" and "edges"."""
    nwbfile = NWBFile(
        session_description="session_description",
        identifier="identifier",
        session_start_time=datetime.datetime.now(datetime.timezone.utc),
    )

    data = np.random.rand(100, 3)  # num_frames x (x, y, z)
    timestamps = np.linspace(0, 10, num=100)  # a timestamp for every frame
    confidence = np.random.rand(100)  # a confidence value for every frame
    front_left_paw = PoseEstimationSeries(
        name="front_left_paw",
        description="Marker placed around fingers of front left paw.",
        data=data,
        unit="pixels",
        reference_frame="(0,0,0) corresponds to ...",
        timestamps=timestamps,
        confidence=confidence,
        confidence_definition="Softmax output of the deep neural network.",
    )

    data = np.random.rand(100, 2)  # num_frames x (x, y)
    timestamps = np.linspace(0, 10, num=100)  # a timestamp for every frame
    confidence = np.random.rand(100)  # a confidence value for every frame
    front_right_paw = PoseEstimationSeries(
        name="front_right_paw",
        description="Marker placed around fingers of front right paw.",
        data=data,
        unit="pixels",
        reference_frame="(0,0,0) corresponds to ...",
        timestamps=front_left_paw,  # link to timestamps of front_left_paw
        confidence=confidence,
        confidence_definition="Softmax output of the deep neural network.",
    )

    pose_estimation_series = [front_left_paw, front_right_paw]

    pe = PoseEstimation(
        pose_estimation_series=pose_estimation_series,
        description="Estimated positions of front paws using DeepLabCut.",
        scorer="DLC_resnet50_openfieldOct30shuffle1_1600",
        source_software="DeepLabCut",
        source_software_version="2.2b8",
        nodes=["front_left_paw", "front_right_paw"],
        edges=np.array([[0, 1]], dtype="uint8"),
    )

    behavior_pm = nwbfile.create_processing_module(name="behavior", description="processed behavioral data")
    behavior_pm.add(pe)

    path = Path(__file__).parent / "0.1.1_poseestimation_nodes_edges.nwb"
    with NWBHDF5IO(path, mode="w") as io:
        io.write(nwbfile)


def generate_pose_estimation_no_cameras():
    """Generate an NWB file with a PoseEstimation object with video metadata and no cameras."""
    nwbfile = NWBFile(
        session_description="session_description",
        identifier="identifier",
        session_start_time=datetime.datetime.now(datetime.timezone.utc),
    )
    pe = PoseEstimation(
        pose_estimation_series=[],
        description="Estimated positions of front paws using DeepLabCut.",
        original_videos=["camera1.mp4", "camera2.mp4"],
        labeled_videos=["camera1_labeled.mp4", "camera2_labeled.mp4"],
        dimensions=np.array([[640, 480], [1024, 768]], dtype="uint8"),
        scorer="DLC_resnet50_openfieldOct30shuffle1_1600",
        source_software="DeepLabCut",
        source_software_version="2.2b8",
    )

    behavior_pm = nwbfile.create_processing_module(name="behavior", description="processed behavioral data")
    behavior_pm.add(pe)

    path = Path(__file__).parent / "0.1.1_poseestimation_no_cameras.nwb"
    with NWBHDF5IO(path, mode="w") as io:
        io.write(nwbfile)


if __name__ == "__main__":
    generate_pose_estimation_nodes_edges()
    generate_pose_estimation_no_cameras()



from pynwb import register_map
from pynwb.io.base import TimeSeriesMap
from pynwb.io.core import NWBContainerMapper

from ..pose import PoseEstimation, PoseEstimationSeries


@register_map(PoseEstimationSeries)
class PoseEstimationSeriesMap(TimeSeriesMap):

    def __init__(self, spec):
        """Map attribute spec "definition" to Python instance attribute "confidence_definition"."""
        super().__init__(spec)
        confidence_spec = self.spec.get_dataset("confidence")
        self.map_spec("confidence_definition", confidence_spec.get_attribute("definition"))


@register_map(PoseEstimation)
class PoseEstimationMap(NWBContainerMapper):

    def __init__(self, spec):
        """Map attribute spec "version" to Python instance attribute "source_software_version"."""
        super().__init__(spec)
        source_software_spec = self.spec.get_dataset("source_software")
        self.map_spec("source_software_version", source_software_spec.get_attribute("version"))

    @NWBContainerMapper.constructor_arg("nodes")
    def nodes(self, builder, manager):
        """Set the constructor arg for 'nodes' to the value of the GroupBuilder dataset "nodes".

        Used when constructing a PoseEstimation container from a written file.

        ndx-pose 0.2.0 introduced a new attribute 'skeleton' to the PoseEstimation container. This Skeleton
        container has two datasets, 'nodes' and 'edges', which were previously stored directly in the
        PoseEstimation container. When data written with ndx-pose versions < 0.2.0 are read, the 'nodes' and
        'edges' arguments in the PoseEstimation constructor are set to the values of the "nodes" and "edges"
        DatasetBuilders read from the file. When data written with ndx-pose versions >= 0.2.0 are read,
        'nodes' and 'edges' are set to None in the PoseEstimation constructor.
        """
        nodes_builder = builder.datasets.get("nodes")
        if nodes_builder:
            nodes = nodes_builder.data
        else:
            nodes = None
        return nodes

    @NWBContainerMapper.constructor_arg("edges")
    def edges(self, builder, manager):
        """Set the constructor arg for 'edges' to the value of the GroupBuilder dataset "edges".

        Used when constructing a PoseEstimation container from a written file.

        ndx-pose 0.2.0 introduced a new attribute 'skeleton' to the PoseEstimation container. This Skeleton
        container has two datasets, 'nodes' and 'edges', which were previously stored directly in the
        PoseEstimation container. When data written with ndx-pose versions < 0.2.0 are read, the 'nodes' and
        'edges' arguments in the PoseEstimation constructor are set to the values of the "nodes" and "edges"
        DatasetBuilders read from the file. When data written with ndx-pose versions >= 0.2.0 are read,
        'nodes' and 'edges' are set to None in the PoseEstimation constructor.
        """
        edges_builder = builder.datasets.get("edges")
        if edges_builder:
            edges = edges_builder.data
        else:
            edges = None
        return edges



# -*- coding: utf-8 -*-
import os.path

from pynwb.spec import (
    export_spec,
    NWBAttributeSpec,
    NWBDatasetSpec,
    NWBGroupSpec,
    NWBLinkSpec,
    NWBNamespaceBuilder,
)


def main():
    # these arguments were auto-generated from your cookiecutter inputs
    ns_builder = NWBNamespaceBuilder(
        doc="NWB extension to store pose estimation data",
        name="ndx-pose",
        version="0.2.0",
        author=[
            "Ryan Ly",
            "Ben Dichter",
            "Alexander Mathis",
            "Liezl Maree",
            "Chris Brozdowski",
            "Heberto Mayorquin",
            "Talmo Pereira",
            "Elizabeth Berrigan",
        ],
        contact=[
            "rly@lbl.gov",
            "bdichter@lbl.gov",
            "alexander.mathis@epfl.ch",
            "lmaree@salk.edu",
            "cbroz@datajoint.com",
            "h.mayorquin@gmail.com",
            "talmo@salk.edu",
            "eberrigan@salk.edu",
        ],
    )

    ns_builder.include_namespace(namespace="core")

    skeleton = NWBGroupSpec(
        neurodata_type_def="Skeleton",
        neurodata_type_inc="NWBDataInterface",
        doc=(
            "Group that holds node and edge data for defining parts of a pose and their connections to one another. "
            "Names should be unique in a file."
        ),
        datasets=[
            NWBDatasetSpec(
                name="nodes",
                doc=(
                    "Array of body part names corresponding to the names of the PoseEstimationSeries objects or "
                    "PoseTraining objects."
                ),
                dtype="text",
                dims=["num_body_parts"],
                shape=[None],
                quantity=1,
            ),
            NWBDatasetSpec(
                name="edges",
                doc=(
                    "Array of pairs of indices corresponding to edges between nodes. Index values correspond to row "
                    "indices of the 'nodes' dataset. Index values use 0-indexing."
                ),
                dtype="uint8",
                dims=["num_edges", "nodes_index, nodes_index"],
                shape=[None, 2],
                quantity="?",
            ),
        ],
        links=[
            NWBLinkSpec(
                doc="The Subject object in the NWB file, if this Skeleton corresponds to the Subject.",
                target_type="Subject",
                quantity="?",
            ),
        ],
    )

    skeletons = NWBGroupSpec(
        neurodata_type_def="Skeletons",
        neurodata_type_inc="NWBDataInterface",
        doc="Organizational group to hold skeletons.",
        default_name="Skeletons",
        groups=[
            NWBGroupSpec(
                neurodata_type_inc="Skeleton",
                doc="Skeleton used in project where each skeleton corresponds to a unique morphology.",
                quantity="*",
            ),
        ],
    )

    pose_estimation_series = NWBGroupSpec(
        neurodata_type_def="PoseEstimationSeries",
        neurodata_type_inc="SpatialSeries",
        doc="Estimated position (x, y) or (x, y, z) of a body part over time.",
        datasets=[
            NWBDatasetSpec(
                name="data",
                doc="Estimated position (x, y) or (x, y, z).",
                dtype="float32",
                dims=[["num_frames", "x, y"], ["num_frames", "x, y, z"]],
                shape=[[None, 2], [None, 3]],
                attributes=[
                    NWBAttributeSpec(
                        name="unit",
                        dtype="text",
                        default_value="pixels",
                        doc=(
                            "Base unit of measurement for working with the data. The default value "
                            "is 'pixels'. Actual stored values are not necessarily stored in these units. "
                            "To access the data in these units, multiply 'data' by 'conversion'."
                        ),
                        required=True,
                    ),
                ],
            ),
            NWBDatasetSpec(
                name="confidence",
                doc="Confidence or likelihood of the estimated positions, scaled to be between 0 and 1.",
                dtype="float32",
                dims=["num_frames"],
                shape=[None],
                attributes=[
                    NWBAttributeSpec(
                        name="definition",
                        dtype="text",
                        doc=(
                            "Description of how the confidence was computed, e.g., "
                            "'Softmax output of the deep neural network'."
                        ),
                        required=False,
                    ),
                ],
            ),
        ],
    )

    pose_estimation = NWBGroupSpec(
        neurodata_type_def="PoseEstimation",
        neurodata_type_inc="NWBDataInterface",
        doc=(
            "Group that holds estimated position data for multiple body parts, computed from the same video with "
            "the same tool/algorithm. The timestamps of each child PoseEstimationSeries type should be the same."
        ),
        default_name="PoseEstimation",
        groups=[
            NWBGroupSpec(
                neurodata_type_inc="PoseEstimationSeries",
                doc="Estimated position data for each body part.",
                quantity="*",
            ),
        ],
        links=[
            NWBLinkSpec(
                doc="Layout of body part locations and connections.",
                target_type="Skeleton",
                quantity="?",
            ),
            NWBLinkSpec(
                target_type="Device",
                doc="Cameras used to record the videos.",
                quantity="*",
            ),
        ],
        datasets=[
            NWBDatasetSpec(
                name="description",
                doc="Description of the pose estimation procedure and output.",
                dtype="text",
                quantity="?",
            ),
            NWBDatasetSpec(
                name="original_videos",
                doc="Paths to the original video files. The number of files should equal the number of camera devices.",
                dtype="text",
                dims=["num_files"],
                shape=[None],
                quantity="?",
            ),
            NWBDatasetSpec(
                name="labeled_videos",
                doc="Paths to the labeled video files. The number of files should equal the number of camera devices.",
                dtype="text",
                dims=["num_files"],
                shape=[None],
                quantity="?",
            ),
            NWBDatasetSpec(
                name="dimensions",
                doc="Dimensions of each labeled video file.",
                dtype="uint8",
                dims=["num_files", "width, height"],
                shape=[None, 2],
                quantity="?",
            ),
            NWBDatasetSpec(
                name="scorer",
                doc="Name of the scorer / algorithm used.",
                dtype="text",
                quantity="?",
            ),
            NWBDatasetSpec(
                name="source_software",
                doc="Name of the software tool used. Specifying the version attribute is strongly encouraged.",
                dtype="text",
                quantity="?",
                attributes=[
                    NWBAttributeSpec(
                        name="version",
                        doc="Version string of the software tool used.",
                        dtype="text",
                        required=False,
                    ),
                ],
            ),
        ],
    )

    skeleton_instance = NWBGroupSpec(
        neurodata_type_def="SkeletonInstance",
        neurodata_type_inc="NWBDataInterface",
        doc="Group that holds ground-truth pose data for a single instance of a skeleton in a single frame.",
        default_name="skeleton_instance",
        links=[
            NWBLinkSpec(
                doc="Layout of body part locations and connections.",
                target_type="Skeleton",
            ),
        ],
        attributes=[  # TODO since the link to the Skeleton is required, is this necessary?
            NWBAttributeSpec(
                name="id",
                doc="ID used to differentiate skeleton instances.",
                dtype="uint8",
                required=False,
            ),
        ],
        datasets=[
            NWBDatasetSpec(
                name="node_locations",
                doc="Locations (x, y) or (x, y, z) of nodes for single instance in single frame.",
                dtype="float",
                dims=[["num_body_parts", "x, y"], ["num_body_parts", "x, y, z"]],
                shape=[[None, 2], [None, 3]],
                quantity=1,
            ),
            NWBDatasetSpec(
                name="node_visibility",
                doc=(
                    "Markers for node visibility where true corresponds to a visible node and false corresponds to "
                    "an occluded node."
                ),
                dtype="bool",
                dims=["num_body_parts"],
                shape=[None],
                quantity="?",
            ),
        ],
    )

    skeleton_instances = NWBGroupSpec(
        neurodata_type_def="SkeletonInstances",
        neurodata_type_inc="NWBDataInterface",
        doc="Organizational group to hold skeleton instances. This is meant to be used within a TrainingFrame.",
        default_name="skeleton_instances",
        groups=[
            NWBGroupSpec(
                neurodata_type_inc="SkeletonInstance",
                doc="Ground-truth position data for a single instance of a skeleton in a single training frame.",
                quantity="*",
            ),
        ],
    )

    source_videos = NWBGroupSpec(
        neurodata_type_def="SourceVideos",
        neurodata_type_inc="NWBDataInterface",
        doc="Organizational group to hold source videos used for training.",
        # this is meant to be used in a PoseTraining object which will enforce this name
        default_name="source_videos",
        groups=[
            NWBGroupSpec(
                neurodata_type_inc="ImageSeries",
                doc="Video of training frames (stored internally or externally",
                quantity="*",
            ),
        ],
    )

    training_frame = NWBGroupSpec(
        neurodata_type_def="TrainingFrame",
        neurodata_type_inc="NWBDataInterface",
        doc="Group that holds ground-truth position data for all instances of a skeleton in a single frame.",
        default_name="TrainingFrame",
        groups=[
            NWBGroupSpec(
                name="skeleton_instances",
                neurodata_type_inc="SkeletonInstances",
                doc="Position data for all instances of a skeleton in a single training frame.",
            ),
        ],
        attributes=[
            NWBAttributeSpec(
                name="annotator",
                doc="Name of annotator who labeled the TrainingFrame.",
                dtype="text",
                required=False,
            ),
            NWBAttributeSpec(
                name="source_video_frame_index",
                doc=(
                    "Frame index of training frame in the original video `source_video`. "
                    "If provided, then `source_video` is required."
                ),
                dtype="uint8",
                required=False,
            ),
            # TODO add inspector check that either both source_video and source_video_frame_index are provided or
            # neither are provided
        ],
        links=[
            NWBLinkSpec(
                name="source_video",
                target_type="ImageSeries",
                doc=(
                    "Link to an ImageSeries representing a video of training frames (stored internally or "
                    "externally). Required if `source_video_frame_index` is provided."
                ),
                quantity="?",
            ),
            NWBLinkSpec(
                name="source_frame",
                target_type="Image",
                doc=(
                    "Link to an internally stored image representing the training frame. The target Image "
                    "should be stored in an Images type in the file."
                ),
                quantity="?",
            ),
        ],
    )

    training_frames = NWBGroupSpec(
        neurodata_type_def="TrainingFrames",
        neurodata_type_inc="NWBDataInterface",
        doc="Organizational group to hold training frames.",
        # this is meant to be used in a PoseTraining object which will enforce this name
        default_name="training_frames",
        groups=[
            NWBGroupSpec(
                neurodata_type_inc="TrainingFrame",
                doc="Ground-truth position data for all instances of a skeleton in a single frame.",
                quantity="*",
            ),
        ],
    )

    pose_training = NWBGroupSpec(
        neurodata_type_def="PoseTraining",
        neurodata_type_inc="NWBDataInterface",
        doc="Group that holds source videos and ground-truth annotations for training a pose estimator.",
        default_name="PoseTraining",
        groups=[
            NWBGroupSpec(
                name="training_frames",
                neurodata_type_inc="TrainingFrames",
                doc="Organizational group to hold training frames.",
                quantity="?",
            ),
            NWBGroupSpec(
                name="source_videos",
                neurodata_type_inc="SourceVideos",
                doc="Organizational group to hold source videos used for training.",
                quantity="?",
            ),
        ],
    )

    new_data_types = [
        skeleton,
        pose_estimation_series,
        pose_estimation,
        training_frame,
        skeleton_instance,
        training_frames,
        skeleton_instances,
        source_videos,
        skeletons,
        pose_training,
    ]

    # export the spec to yaml files in the spec folder
    output_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", "..", "spec"))
    export_spec(ns_builder, new_data_types, output_dir)
    print("Spec files generated. Please make sure to rerun `pip install .` to load the changes.")


if __name__ == "__main__":
    # usage: python create_extension_spec.py
    main()



from typing import Optional, Any, Union, List

import numpy as np
from pynwb import NWBFile
from pynwb.image import ImageSeries, Image, RGBImage
from pynwb.testing.mock.utils import name_generator
from pynwb.testing.mock.device import mock_Device

from ...pose import (
    PoseEstimationSeries,
    Skeleton,
    PoseEstimation,
    SkeletonInstance,
    SkeletonInstances,
    TrainingFrame,
    Skeletons,
)


def mock_PoseEstimationSeries(
    *,
    name: Optional[str] = None,
    description: Optional[str] = "A description.",
    data=None,
    unit: Optional[str] = "pixels",
    resolution: float = -1.0,
    conversion: float = 1.0,
    offset: float = 0.0,
    timestamps=None,
    starting_time: Optional[float] = None,
    rate: Optional[float] = None,
    reference_frame: Optional[str] = "(0,0,0) corresponds to ...",
    confidence=None,
    confidence_definition: Optional[str] = "Softmax output of the deep neural network.",
):
    if data is None:
        data = np.arange(30, dtype=np.float64).reshape((10, 3))
    if timestamps is not None:
        rate = None
    if timestamps is None and rate is None:
        timestamps = np.linspace(0, 10, num=len(data))  # a timestamp for every frame
    if confidence is None:
        confidence = np.linspace(0, 1, num=len(data))  # confidence value for every frame
    pes = PoseEstimationSeries(
        name=name or name_generator("PoseEstimationSeries"),
        data=data,
        unit=unit,
        resolution=resolution,
        conversion=conversion,
        offset=offset,
        timestamps=timestamps,
        starting_time=starting_time,
        rate=rate,
        reference_frame=reference_frame,
        description=description,
        confidence=confidence,
        confidence_definition=confidence_definition,
    )

    return pes


def mock_Skeleton(
    *,
    name: Optional[str] = None,
    nodes: Optional[Any] = None,
    edges: Optional[Any] = None,
):
    if nodes is None and edges is None:
        nodes = ["node1", "node2", "node3"]
        edges = np.array([(0, 1), (1, 2)], dtype="uint8")
    skeleton = Skeleton(
        name=name or name_generator("Skeleton"),
        nodes=nodes,
        edges=edges,
    )

    return skeleton


def mock_PoseEstimation(
    *,
    nwbfile: NWBFile,
    pose_estimation_series: Optional[list] = None,
    skeleton: Optional[Skeleton] = None,
    devices: Optional[list] = None,
    description: Optional[str] = "Estimated positions of front paws using DeepLabCut.",
    original_videos: Optional[list] = None,
    labeled_videos: Optional[list] = None,
    dimensions: Optional[np.ndarray] = None,
    scorer: Optional[str] = "DLC_resnet50_openfieldOct30shuffle1_1600",
    source_software: Optional[str] = "DeepLabCut",
    source_software_version: Optional[str] = "2.2b8",
):
    """Create a mock PoseEstimation object.

    NWBFile should be provided so that the skeleton can be added to the NWBFile in a PoseTraining object.
    """
    skeleton = skeleton or mock_Skeleton()
    pose_estimation_series = pose_estimation_series or [mock_PoseEstimationSeries(name=name) for name in skeleton.nodes]
    pe = PoseEstimation(
        pose_estimation_series=pose_estimation_series,
        description=description,
        original_videos=original_videos or ["camera1.mp4"],
        labeled_videos=labeled_videos or ["camera1_labeled.mp4"],
        dimensions=dimensions or np.array([[640, 480]], dtype="uint16"),
        devices=devices or [mock_Device(nwbfile=nwbfile)],
        scorer=scorer,
        source_software=source_software,
        source_software_version=source_software_version,
        skeleton=skeleton,
    )

    if nwbfile is not None:
        skeletons = Skeletons(skeletons=[skeleton])

        if "behavior" not in nwbfile.processing:
            behavior_pm = nwbfile.create_processing_module(name="behavior", description="processed behavioral data")
        else:
            behavior_pm = nwbfile.processing["behavior"]
        behavior_pm.add(pe)
        behavior_pm.add(skeletons)

    return pe


def mock_SkeletonInstance(
    *,
    name: Optional[str] = None,
    id: Optional[np.uint] = np.uint(10),
    node_locations: Optional[Any] = None,
    node_visibility: Optional[list] = None,
    skeleton: Optional[Skeleton] = None,
):
    if node_locations is None and node_visibility is None:
        num_nodes = 3
    elif node_locations is not None:
        num_nodes = len(node_locations)
    else:
        num_nodes = len(node_visibility)

    if skeleton is None:
        skeleton = mock_Skeleton(
            nodes=["node" + str(i) for i in range(num_nodes)],
            edges=np.array([[0, 1]], dtype="uint8"),
        )
    if node_locations is None:
        node_locations = np.arange(num_nodes * 2, dtype=np.float64).reshape((num_nodes, 2))

    if name is None:
        name = skeleton.name + "_instance_" + str(id)
    if node_visibility is None:
        node_visibility = np.ones(num_nodes, dtype="bool")

    skeleton_instance = SkeletonInstance(
        name=name,
        id=id,
        node_locations=node_locations,
        node_visibility=node_visibility,
        skeleton=skeleton,
    )

    return skeleton_instance


def mock_SkeletonInstances(skeleton_instances: Union[SkeletonInstance, List[SkeletonInstance]] = None):
    if skeleton_instances is None:
        skeleton_instances = [mock_SkeletonInstance()]
    if not isinstance(skeleton_instances, list):
        skeleton_instances = [skeleton_instances]
    return SkeletonInstances(
        skeleton_instances=skeleton_instances,
    )


def mock_source_video(
    *,
    name: Optional[str] = None,
):
    source_video = ImageSeries(
        name=name or name_generator("ImageSeries"),
        description="Training video used to train the pose estimation model.",
        unit="NA",
        format="external",
        external_file=["path/to/camera1.mp4"],
        dimension=[640, 480],
        starting_frame=[0],
        rate=30.0,
    )
    return source_video


def mock_source_frame(
    *,
    name: Optional[str] = None,
):
    return RGBImage(name=name, data=np.random.rand(640, 480, 3).astype("uint8"))


def mock_TrainingFrame(
    *,
    name: Optional[str] = None,
    annotator: Optional[str] = "Awesome Possum",
    skeleton_instances: Optional[SkeletonInstances] = None,
    source_video: Optional[ImageSeries] = None,
    source_frame: Optional[Image] = None,
    source_video_frame_index: Optional[np.uint] = np.uint(10),
):
    training_frame = TrainingFrame(
        name=name or name_generator("TrainingFrame"),
        annotator=annotator,
        skeleton_instances=skeleton_instances or mock_SkeletonInstances(),
        source_video=source_video or (mock_source_video() if source_frame is None else None),
        source_frame=source_frame,
        source_video_frame_index=source_video_frame_index,
    )
    return training_frame



import datetime
import numpy as np

from pynwb import NWBFile
from pynwb.device import Device
from pynwb.testing import TestCase
from pynwb.file import Subject

from ndx_pose import (
    PoseEstimationSeries,
    Skeleton,
    PoseEstimation,
    TrainingFrame,
    SkeletonInstance,
    PoseTraining,
    TrainingFrames,
    SourceVideos,
)
from ndx_pose.testing.mock.pose import (
    mock_PoseEstimationSeries,
    mock_SkeletonInstances,
    mock_source_video,
    mock_source_frame,
    mock_Skeleton,
    mock_SkeletonInstance,
    mock_TrainingFrame,
)

# NOTE Skeletons, TrainingFrames, SourceVideos are tested within PoseTraining but not separately tested


class TestPoseEstimationSeriesConstructor(TestCase):
    def test_constructor(self):
        data = np.random.rand(100, 3)  # num_frames x (x, y, z)
        timestamps = np.linspace(0, 10, num=100)  # a timestamp for every frame
        confidence = np.random.rand(100)  # a confidence value for every frame
        pes = PoseEstimationSeries(
            name="front_left_paw",
            description="Marker placed around fingers of front left paw.",
            data=data,
            unit="pixels",
            reference_frame="(0,0,0) corresponds to ...",
            timestamps=timestamps,
            confidence=confidence,
            confidence_definition="Softmax output of the deep neural network.",
        )

        self.assertEqual(pes.name, "front_left_paw")
        self.assertEqual(pes.description, "Marker placed around fingers of front left paw.")
        np.testing.assert_array_equal(pes.data, data)
        self.assertEqual(pes.unit, "pixels")
        self.assertEqual(pes.reference_frame, "(0,0,0) corresponds to ...")
        np.testing.assert_array_equal(pes.timestamps, timestamps)
        np.testing.assert_array_equal(pes.confidence, confidence)
        self.assertEqual(pes.confidence_definition, "Softmax output of the deep neural network.")


class TestSkeleton(TestCase):
    def test_init(self):
        subject = Subject(subject_id="MOUSE001", species="Mus musculus")
        skeleton = Skeleton(
            name="subject1",
            nodes=["front_left_paw", "body", "front_right_paw"],
            # edge between front left paw and body, edge between body and front right paw.
            # the values are the indices of the nodes in the nodes list.
            edges=np.array([[0, 1], [1, 2]], dtype="uint8"),
            subject=subject,
        )
        self.assertEqual(skeleton.name, "subject1")
        self.assertEqual(skeleton.nodes, ["front_left_paw", "body", "front_right_paw"])
        np.testing.assert_array_equal(skeleton.edges, np.array([[0, 1], [1, 2]], dtype="uint8"))
        self.assertIs(skeleton.subject, subject)

    def test_init_no_subject(self):
        skeleton = Skeleton(
            name="subject1",
            nodes=["front_left_paw", "body", "front_right_paw"],
            edges=np.array([[0, 1], [1, 2]], dtype="uint8"),
        )
        self.assertIsNone(skeleton.subject)


class TestPoseEstimationConstructor(TestCase):
    def setUp(self):
        nwbfile = NWBFile(
            session_description="session_description",
            identifier="identifier",
            session_start_time=datetime.datetime.now(datetime.timezone.utc),
        )
        nwbfile.create_device(name="camera1")
        nwbfile.create_device(name="camera2")

        self.nwbfile = nwbfile

    def test_constructor(self):
        """Test that the constructor for PoseEstimation sets values as expected."""
        front_left_paw = mock_PoseEstimationSeries(
            name="front_left_paw",
        )
        body = mock_PoseEstimationSeries(
            name="body",
            timestamps=front_left_paw,  # link to timestamps of front_left_paw
        )
        front_right_paw = mock_PoseEstimationSeries(
            name="front_right_paw",
            timestamps=front_left_paw,  # link to timestamps of front_left_paw
        )
        pose_estimation_series = [front_left_paw, body, front_right_paw]
        skeleton = mock_Skeleton(
            nodes=["front_left_paw", "body", "front_right_paw"],
            # edge between front left paw and body, edge between body and front right paw.
            # the values are the indices of the nodes in the nodes list.
            edges=np.array([[0, 1], [1, 2]], dtype="uint8"),
        )

        pe = PoseEstimation(
            pose_estimation_series=pose_estimation_series,
            description="Estimated positions of front paws using DeepLabCut.",
            original_videos=["camera1.mp4", "camera2.mp4"],
            labeled_videos=["camera1_labeled.mp4", "camera2_labeled.mp4"],
            dimensions=np.array([[640, 480], [1024, 768]], dtype="uint16"),
            devices=[self.nwbfile.devices["camera1"], self.nwbfile.devices["camera2"]],
            scorer="DLC_resnet50_openfieldOct30shuffle1_1600",
            source_software="DeepLabCut",
            source_software_version="2.2b8",
            skeleton=skeleton,
        )

        self.assertEqual(pe.name, "PoseEstimation")
        self.assertEqual(len(pe.pose_estimation_series), 3)
        self.assertIs(pe.pose_estimation_series["front_left_paw"], pose_estimation_series[0])
        self.assertIs(pe.pose_estimation_series["body"], pose_estimation_series[1])
        self.assertIs(pe.pose_estimation_series["front_right_paw"], pose_estimation_series[2])
        self.assertEqual(pe.description, "Estimated positions of front paws using DeepLabCut.")
        self.assertEqual(pe.original_videos, ["camera1.mp4", "camera2.mp4"])
        self.assertEqual(pe.labeled_videos, ["camera1_labeled.mp4", "camera2_labeled.mp4"])
        np.testing.assert_array_equal(pe.dimensions, np.array([[640, 480], [1024, 768]], dtype="uint16"))
        self.assertEqual(len(pe.devices), 2)
        self.assertIs(pe.devices[0], self.nwbfile.devices["camera1"])
        self.assertIs(pe.devices[1], self.nwbfile.devices["camera2"])
        self.assertEqual(pe.scorer, "DLC_resnet50_openfieldOct30shuffle1_1600")
        self.assertEqual(pe.source_software, "DeepLabCut")
        self.assertEqual(pe.source_software_version, "2.2b8")
        self.assertIs(pe.skeleton, skeleton)

    def test_bad_device_link(self):
        front_left_paw = mock_PoseEstimationSeries(
            name="front_left_paw",
        )
        body = mock_PoseEstimationSeries(
            name="body",
            timestamps=front_left_paw,  # link to timestamps of front_left_paw
        )
        front_right_paw = mock_PoseEstimationSeries(
            name="front_right_paw",
            timestamps=front_left_paw,  # link to timestamps of front_left_paw
        )
        pose_estimation_series = [front_left_paw, body, front_right_paw]
        skeleton = mock_Skeleton(
            nodes=["front_left_paw", "body", "front_right_paw"],
            # edge between front left paw and body, edge between body and front right paw.
            # the values are the indices of the nodes in the nodes list.
            edges=np.array([[0, 1], [1, 2]], dtype="uint8"),
        )
        camera1 = Device(name="camera1")
        camera2 = Device(name="camera2")

        msg = "All devices linked to from a PoseEstimation object must be added to the NWBFile first."
        with self.assertRaisesWith(ValueError, msg):
            PoseEstimation(
                pose_estimation_series=pose_estimation_series,
                description="Estimated positions of front paws using DeepLabCut.",
                original_videos=["camera1.mp4", "camera2.mp4"],
                labeled_videos=["camera1_labeled.mp4", "camera2_labeled.mp4"],
                dimensions=np.array([[640, 480], [1024, 768]], dtype="uint16"),
                devices=[camera1, camera2],
                scorer="DLC_resnet50_openfieldOct30shuffle1_1600",
                source_software="DeepLabCut",
                source_software_version="2.2b8",
                skeleton=skeleton,
            )

    def test_constructor_nodes_edges(self):
        """Test the old constructor for PoseEstimation with nodes and edges."""
        front_left_paw = mock_PoseEstimationSeries(
            name="front_left_paw",
        )
        body = mock_PoseEstimationSeries(
            name="body",
            timestamps=front_left_paw,  # link to timestamps of front_left_paw
        )
        front_right_paw = mock_PoseEstimationSeries(
            name="front_right_paw",
            timestamps=front_left_paw,  # link to timestamps of front_left_paw
        )
        pose_estimation_series = [front_left_paw, body, front_right_paw]

        msg = (
            "The 'nodes' and 'edges' constructor arguments are deprecated. Please use the 'skeleton' argument instead. "
            "These will be removed in a future release."
        )
        with self.assertWarnsWith(DeprecationWarning, msg):
            pe = PoseEstimation(
                pose_estimation_series=pose_estimation_series,
                description="Estimated positions of front paws using DeepLabCut.",
                original_videos=["camera1.mp4", "camera2.mp4"],
                labeled_videos=["camera1_labeled.mp4", "camera2_labeled.mp4"],
                dimensions=np.array([[640, 480], [1024, 768]], dtype="uint16"),
                devices=[
                    self.nwbfile.devices["camera1"],
                    self.nwbfile.devices["camera2"],
                ],
                scorer="DLC_resnet50_openfieldOct30shuffle1_1600",
                source_software="DeepLabCut",
                source_software_version="2.2b8",
                nodes=["front_left_paw", "body", "front_right_paw"],
                edges=np.array([[0, 1], [1, 2]], dtype="uint8"),
            )
        self.assertEqual(pe.nodes, ["front_left_paw", "body", "front_right_paw"])
        np.testing.assert_array_equal(pe.edges, np.array([[0, 1], [1, 2]], dtype="uint8"))
        skeleton = Skeleton(
            name="subject",
            nodes=["front_left_paw", "body", "front_right_paw"],
            edges=np.array([[0, 1], [1, 2]], dtype="uint8"),
        )
        self.assertEqual(pe.skeleton.name, skeleton.name)
        self.assertEqual(pe.skeleton.nodes, skeleton.nodes)
        np.testing.assert_array_equal(pe.skeleton.edges, skeleton.edges)


class TestSkeletonInstance(TestCase):
    def test_constructor(self):
        skeleton = mock_Skeleton(
            nodes=["left_eye", "right_eye"],
            edges=np.array([[0, 1]], dtype="uint8"),
        )

        node_locations = np.array(
            [
                [1, 5],
                [20, 30],
            ]
        )
        instance = SkeletonInstance(
            id=np.uint(10),
            node_locations=node_locations,
            node_visibility=[
                True,
                False,
            ],
            skeleton=skeleton,
        )
        self.assertEqual(instance.id, np.uint(10))
        np.testing.assert_array_equal(instance.node_locations, node_locations)
        self.assertEqual(instance.node_visibility, [True, False])
        self.assertIs(instance.skeleton, skeleton)


class TestTrainingFrame(TestCase):
    def test_constructor(self):
        skeleton_instances = mock_SkeletonInstances()
        source_video = mock_source_video(name="source_video")
        training_frame = TrainingFrame(
            name="frame0",
            annotator="Awesome Possum",
            skeleton_instances=skeleton_instances,
            source_video=source_video,
            source_video_frame_index=np.uint(0),
        )
        self.assertEqual(training_frame.name, "frame0")
        self.assertEqual(training_frame.annotator, "Awesome Possum")
        self.assertIs(training_frame.skeleton_instances, skeleton_instances)
        self.assertIs(training_frame.source_video, source_video)
        self.assertEqual(training_frame.source_video_frame_index, np.uint(0))


class TestTrainingFrameImage(TestCase):
    def test_constructor(self):
        skeleton_instances = mock_SkeletonInstances()
        source_frame = mock_source_frame(name="frame0_image")
        training_frame = TrainingFrame(
            name="frame0",
            annotator="Awesome Possum",
            skeleton_instances=skeleton_instances,
            source_frame=source_frame,
            source_video_frame_index=np.uint(0),
        )
        self.assertEqual(training_frame.name, "frame0")
        self.assertEqual(training_frame.annotator, "Awesome Possum")
        self.assertIs(training_frame.skeleton_instances, skeleton_instances)
        self.assertIs(training_frame.source_frame, source_frame)
        self.assertEqual(training_frame.source_video_frame_index, np.uint(0))


class TestPoseTraining(TestCase):
    def test_constructor(self):
        skeleton1 = mock_Skeleton(name="subject1")
        skeleton2 = mock_Skeleton(name="subject2")
        source_video = mock_source_video(name="source_video")
        sk1_instance10 = mock_SkeletonInstance(id=np.uint(10), skeleton=skeleton1)
        sk1_instance11 = mock_SkeletonInstance(id=np.uint(11), skeleton=skeleton1)
        sk1_instances = mock_SkeletonInstances(skeleton_instances=[sk1_instance10, sk1_instance11])
        sk1_training_frame = mock_TrainingFrame(
            name="skeleton1_frame10",
            skeleton_instances=sk1_instances,
            source_video=source_video,
            source_video_frame_index=np.uint(10),
        )
        sk2_instance10 = mock_SkeletonInstance(id=np.uint(10), skeleton=skeleton2)
        sk2_instance11 = mock_SkeletonInstance(id=np.uint(11), skeleton=skeleton2)
        sk2_instance12 = mock_SkeletonInstance(id=np.uint(12), skeleton=skeleton2)
        sk2_instances = mock_SkeletonInstances(skeleton_instances=[sk2_instance10, sk2_instance11, sk2_instance12])
        sk2_training_frame = mock_TrainingFrame(
            name="skeleton2_frame10",
            skeleton_instances=sk2_instances,
            source_video=source_video,
            source_video_frame_index=np.uint(10),
        )

        training_frames = TrainingFrames(training_frames=[sk1_training_frame, sk2_training_frame])
        source_videos = SourceVideos(image_series=[source_video])

        pose_training = PoseTraining(
            training_frames=training_frames,
            source_videos=source_videos,
        )
        self.assertEqual(len(pose_training.training_frames.training_frames), 2)
        self.assertIs(
            pose_training.training_frames.training_frames["skeleton1_frame10"],
            sk1_training_frame,
        )
        self.assertIs(
            pose_training.training_frames.training_frames["skeleton2_frame10"],
            sk2_training_frame,
        )
        self.assertEqual(len(pose_training.source_videos.image_series), 1)
        self.assertIs(pose_training.source_videos.image_series["source_video"], source_video)


class TestPoseTrainingImages(TestCase):
    def test_constructor(self):
        skeleton1 = mock_Skeleton(name="subject1")
        skeleton2 = mock_Skeleton(name="subject2")

        source_frame_10 = mock_source_frame(name="source_frame_10")
        sk1_instance10 = mock_SkeletonInstance(id=np.uint(10), skeleton=skeleton1)
        sk1_instance11 = mock_SkeletonInstance(id=np.uint(11), skeleton=skeleton1)
        sk1_instances = mock_SkeletonInstances(skeleton_instances=[sk1_instance10, sk1_instance11])
        sk1_training_frame = mock_TrainingFrame(
            name="frame10",
            skeleton_instances=sk1_instances,
            source_frame=source_frame_10,
            source_video_frame_index=np.uint(10),
        )

        source_frame_11 = mock_source_frame(name="source_frame_11")

        sk2_instance10 = mock_SkeletonInstance(id=np.uint(10), skeleton=skeleton2)
        sk2_instance11 = mock_SkeletonInstance(id=np.uint(11), skeleton=skeleton2)
        sk2_instance12 = mock_SkeletonInstance(id=np.uint(12), skeleton=skeleton2)
        sk2_instances = mock_SkeletonInstances(skeleton_instances=[sk2_instance10, sk2_instance11, sk2_instance12])
        sk2_training_frame = mock_TrainingFrame(
            name="frame11",
            skeleton_instances=sk2_instances,
            source_frame=source_frame_11,
            source_video_frame_index=np.uint(11),
        )

        training_frames = TrainingFrames(training_frames=[sk1_training_frame, sk2_training_frame])

        pose_training = PoseTraining(
            training_frames=training_frames,
        )
        self.assertEqual(len(pose_training.training_frames.training_frames), 2)
        self.assertIs(pose_training.training_frames.training_frames["frame10"], sk1_training_frame)
        self.assertIs(pose_training.training_frames.training_frames["frame11"], sk2_training_frame)
        self.assertIsNone(pose_training.source_videos)



"""Evaluate examples of how to use the ndx-pose extension."""

import subprocess
from pathlib import Path

def test_example_usage_estimates_only():
    """Call examples/write_pose_estimates_only.py and check that it runs without errors."""
    subprocess.run(["python", "examples/write_pose_estimates_only.py"], check=True)

    # Remove the generated test_pose.nwb if it exists
    if Path("test_pose.nwb").exists():
        Path("test_pose.nwb").unlink()


def test_example_usage_training_only():
    """Call examples/write_pose_training.py and check that it runs without errors."""
    subprocess.run(["python", "examples/write_pose_training.py"], check=True)

    # Remove the generated test_pose.nwb if it exists
    if Path("test_pose.nwb").exists():
        Path("test_pose.nwb").unlink()



import datetime
import numpy as np
from pynwb import NWBFile, NWBHDF5IO
from pynwb.file import Subject
from pynwb.image import ImageSeries
from ndx_pose import (
    Skeleton,
    SkeletonInstance,
    TrainingFrame,
    PoseTraining,
    Skeletons,
    TrainingFrames,
    SourceVideos,
    SkeletonInstances,
)

# initialize an NWBFile object
nwbfile = NWBFile(
    session_description="session_description",
    identifier="identifier",
    session_start_time=datetime.datetime.now(datetime.timezone.utc),
)

# add a subject to the NWB file
subject = Subject(subject_id="subject1", species="Mus musculus")
nwbfile.subject = subject

# in this example, we have two subjects in the training data and therefore two skeletons.
# each skeleton defines the relationship between the markers.
# Skeleton names must be unique because the Skeleton objects will be added to a Skeletons container object
# which requires unique names.
skeleton1 = Skeleton(
    name="subject1_skeleton",
    nodes=["front_left_paw", "body", "front_right_paw"],
    # edge between front left paw and body, edge between body and front right paw.
    # the values are the indices of the nodes in the nodes list.
    edges=np.array([[0, 1], [1, 2]], dtype="uint8"),
)
skeleton2 = Skeleton(
    name="subject2_skeleton",
    nodes=["front_left_paw", "body", "front_right_paw"],
    # edge between front left paw and body, edge between body and front right paw.
    # the values are the indices of the nodes in the nodes list.
    edges=np.array([[0, 1], [1, 2]], dtype="uint8"),
)

skeletons = Skeletons(skeletons=[skeleton1, skeleton2])

# create a device for the camera
camera1 = nwbfile.create_device(
    name="camera1",
    description="camera for recording behavior",
    manufacturer="my manufacturer",
)

# next, we specify the ground truth data that was used to train the pose estimation model.
# this includes the training video and the ground truth annotations for each frame.

# create an ImageSeries that represents the raw video that was used to train the pose estimation model.
# the video can be stored as an MP4 file that is linked to from this ImageSeries object.
# if there are multiple videos, the names must be unique because they will be added to a SourceVideos
# container object which requires unique names.
training_video1 = ImageSeries(
    name="source_video",
    description="Training video used to train the pose estimation model.",
    unit="NA",
    format="external",
    external_file=["path/to/camera1.mp4"],
    dimension=[640, 480],
    starting_frame=[0],
    rate=30.0,
)

# initial locations ((x, y) coordinates) of each node in the skeleton.
# the order of the nodes is defined by the skeleton.
node_locations_sk1 = np.array(
    [
        [10.0, 10.0],  # front_left_paw
        [20.0, 20.0],  # body
        [30.0, 10.0],  # front_right_paw
    ]
)
node_locations_sk2 = np.array(
    [
        [40.0, 40.0],  # front_left_paw
        [50.0, 50.0],  # body
        [60.0, 60.0],  # front_right_paw
    ]
)

# in this example, frame indices 0, 5, 10, ..., 500 from the training video were used for training.
# each training frame has two skeleton instances, one for each skeleton.
training_frames_list = []
for i in range(0, 500, 5):
    skeleton_instances_list = []
    # add some noise to the node locations from the previous frame
    node_locations_sk1 = node_locations_sk1 + np.random.rand(3, 2)
    instance_sk1 = SkeletonInstance(
        name="skeleton1_instance",
        id=np.uint(i),
        node_locations=node_locations_sk1,
        node_visibility=[
            True,  # front_left_paw
            True,  # body
            True,  # front_right_paw
        ],
        skeleton=skeleton1,  # link to the skeleton
    )
    skeleton_instances_list.append(instance_sk1)

    # add some noise to the node locations from the previous frame
    node_locations_sk2 = node_locations_sk2 + np.random.rand(3, 2)
    instance_sk2 = SkeletonInstance(
        name="skeleton2_instance",
        id=np.uint(i),
        node_locations=node_locations_sk2,
        node_visibility=[
            True,  # front_left_paw
            True,  # body
            True,  # front_right_paw
        ],
        skeleton=skeleton2,  # link to the skeleton
    )
    skeleton_instances_list.append(instance_sk2)

    # store the skeleton instances in a SkeletonInstances object
    skeleton_instances = SkeletonInstances(skeleton_instances=skeleton_instances_list)

    # TrainingFrame names must be unique because the TrainingFrame objects will be added to a
    # TrainingFrames container object which requires unique names.
    # the source video frame index is the index of the frame in the source video, which is useful
    # for linking the training frames to the source video.
    training_frame = TrainingFrame(
        name=f"frame_{i}",
        annotator="Bilbo Baggins",
        skeleton_instances=skeleton_instances,
        source_video=training_video1,
        source_video_frame_index=np.uint(i),
    )
    training_frames_list.append(training_frame)

# store the training frames and source videos in their corresponding container objects
training_frames = TrainingFrames(training_frames=training_frames_list)
source_videos = SourceVideos(image_series=[training_video1])

# store the skeletons group, training frames group, and source videos group in a PoseTraining object
pose_training = PoseTraining(
    training_frames=training_frames,
    source_videos=source_videos,
)

# create a "behavior" processing module to store the PoseEstimation and PoseTraining objects
behavior_pm = nwbfile.create_processing_module(
    name="behavior",
    description="processed behavioral data",
)
behavior_pm.add(skeletons)
behavior_pm.add(pose_training)

# write the NWBFile to disk
path = "test_pose.nwb"
with NWBHDF5IO(path, mode="w") as io:
    io.write(nwbfile)

# read the NWBFile from disk and print out the PoseEstimation and PoseTraining objects
# as well as the first training frame
with NWBHDF5IO(path, mode="r") as io:
    read_nwbfile = io.read()
    print(read_nwbfile.processing["behavior"]["Skeletons"]["subject1_skeleton"])
    read_pt = read_nwbfile.processing["behavior"]["PoseTraining"]
    print(read_pt.training_frames["frame_10"].skeleton_instances["skeleton2_instance"].node_locations[:])



import datetime
import numpy as np

from pynwb import NWBHDF5IO, NWBFile
from pynwb.testing import TestCase, remove_test_file, NWBH5IOFlexMixin

from ndx_pose import (
    PoseEstimationSeries,
    PoseEstimation,
    PoseTraining,
    Skeletons,
    SourceVideos,
    TrainingFrames,
)
from ndx_pose.testing.mock.pose import (
    mock_PoseEstimationSeries,
    mock_Skeleton,
    mock_PoseEstimation,
    mock_SkeletonInstance,
    mock_SkeletonInstances,
    mock_TrainingFrame,
    mock_source_video,
)


class TestPoseEstimationSeriesRoundtrip(TestCase):
    """Simple roundtrip test for PoseEstimationSeries."""

    def setUp(self):
        self.nwbfile = NWBFile(
            session_description="session_description",
            identifier="identifier",
            session_start_time=datetime.datetime.now(datetime.timezone.utc),
        )
        self.path = "test_pose.nwb"

    def tearDown(self):
        remove_test_file(self.path)

    def test_roundtrip(self):
        """
        Add a PoseEstimationSeries to an NWBFile, write it, read it, and test that the read object matches the original.
        """
        data = np.random.rand(100, 3)  # num_frames x (x, y, z)
        timestamps = np.linspace(0, 10, num=100)  # a timestamp for every frame
        confidence = np.random.rand(100)  # a confidence value for every frame
        pes = PoseEstimationSeries(
            name="front_left_paw",
            description="Marker placed around fingers of front left paw.",
            data=data,
            unit="pixels",
            reference_frame="(0,0,0) corresponds to ...",
            timestamps=timestamps,
            confidence=confidence,
            confidence_definition="Softmax output of the deep neural network.",
        )

        # ideally the PoseEstimationSeries is added to a PoseEstimation object but here, test just the series
        behavior_pm = self.nwbfile.create_processing_module(name="behavior", description="processed behavioral data")
        behavior_pm.add(pes)

        with NWBHDF5IO(self.path, mode="w") as io:
            io.write(self.nwbfile)

        with NWBHDF5IO(self.path, mode="r", load_namespaces=True) as io:
            read_nwbfile = io.read()
            self.assertContainerEqual(pes, read_nwbfile.processing["behavior"]["front_left_paw"])

    def test_roundtrip_link_timestamps(self):
        """
        Test roundtrip of two PoseEstimationSeries where the timestamps of one links to the timestamps of the other.
        """
        data = np.random.rand(100, 3)  # num_frames x (x, y, z)
        timestamps = np.linspace(0, 10, num=100)  # a timestamp for every frame
        confidence = np.random.rand(100)  # a confidence value for every frame
        front_left_paw = PoseEstimationSeries(
            name="front_left_paw",
            description="Marker placed around fingers of front left paw.",
            data=data,
            unit="pixels",
            reference_frame="(0,0,0) corresponds to ...",
            timestamps=timestamps,
            confidence=confidence,
            confidence_definition="Softmax output of the deep neural network.",
        )

        data = np.random.rand(100, 3)  # num_frames x (x, y, z)
        confidence = np.random.rand(100)  # a confidence value for every frame
        front_right_paw = PoseEstimationSeries(
            name="front_right_paw",
            description="Marker placed around fingers of front left paw.",
            data=data,
            unit="pixels",
            reference_frame="(0,0,0) corresponds to ...",
            timestamps=front_left_paw,  # link to timestamps of front_left_paw
            confidence=confidence,
            confidence_definition="Softmax output of the deep neural network.",
        )

        # ideally the PoseEstimationSeries is added to a PoseEstiamtion object but here, test just the series
        behavior_pm = self.nwbfile.create_processing_module(name="behavior", description="processed behavioral data")
        behavior_pm.add(front_left_paw)
        behavior_pm.add(front_right_paw)

        with NWBHDF5IO(self.path, mode="w") as io:
            io.write(self.nwbfile)

        with NWBHDF5IO(self.path, mode="r", load_namespaces=True) as io:
            read_nwbfile = io.read()
            self.assertContainerEqual(front_left_paw, read_nwbfile.processing["behavior"]["front_left_paw"])
            self.assertContainerEqual(front_right_paw, read_nwbfile.processing["behavior"]["front_right_paw"])
            self.assertIs(
                read_nwbfile.processing["behavior"]["front_left_paw"].timestamps,
                read_nwbfile.processing["behavior"]["front_right_paw"].timestamps,
            )


class TestPoseEstimationSeriesRoundtripPyNWB(NWBH5IOFlexMixin, TestCase):
    """Complex, more complete roundtrip test for PoseEstimationSeries using pynwb.testing infrastructure."""

    def getContainerType(self):
        return "PoseEstimationSeries"

    def addContainer(self):
        """Add the test PoseEstimationSeries to the given NWBFile"""
        pes = mock_PoseEstimationSeries(name="test_PES")

        behavior_pm = self.nwbfile.create_processing_module(name="behavior", description="processed behavioral data")
        behavior_pm.add(pes)

    def getContainer(self, nwbfile: NWBFile):
        return nwbfile.processing["behavior"]["test_PES"]


class TestPoseEstimationRoundtrip(TestCase):
    """Simple roundtrip test for PoseEstimation."""

    def setUp(self):
        self.nwbfile = NWBFile(
            session_description="session_description",
            identifier="identifier",
            session_start_time=datetime.datetime.now(datetime.timezone.utc),
        )
        self.nwbfile.create_device(name="camera1")
        self.nwbfile.create_device(name="camera2")
        self.path = "test_pose.nwb"

    def tearDown(self):
        remove_test_file(self.path)

    def test_roundtrip(self):
        """
        Add a PoseEstimation to an NWBFile, write it, read it, and test that the read object matches the original.
        """
        skeleton = mock_Skeleton()
        skeletons = Skeletons(skeletons=[skeleton])

        pose_estimation_series = [mock_PoseEstimationSeries(name=name) for name in skeleton.nodes]
        pe = PoseEstimation(
            pose_estimation_series=pose_estimation_series,
            description="Estimated positions of front paws using DeepLabCut.",
            original_videos=["camera1.mp4"],
            labeled_videos=["camera1_labeled.mp4"],
            dimensions=np.array([[640, 480]], dtype="uint16"),
            devices=[self.nwbfile.devices["camera1"]],
            scorer="DLC_resnet50_openfieldOct30shuffle1_1600",
            source_software="DeepLabCut",
            source_software_version="2.2b8",
            skeleton=skeleton,
        )

        behavior_pm = self.nwbfile.create_processing_module(name="behavior", description="processed behavioral data")
        behavior_pm.add(pe)
        behavior_pm.add(skeletons)

        with NWBHDF5IO(self.path, mode="w") as io:
            io.write(self.nwbfile)

        with NWBHDF5IO(self.path, mode="r", load_namespaces=True) as io:
            read_nwbfile = io.read()
            read_pe = read_nwbfile.processing["behavior"]["PoseEstimation"]
            self.assertContainerEqual(read_pe, pe)
            self.assertEqual(len(read_pe.pose_estimation_series), 3)
            self.assertContainerEqual(read_pe.pose_estimation_series["node1"], pose_estimation_series[0])
            self.assertContainerEqual(read_pe.pose_estimation_series["node2"], pose_estimation_series[1])
            self.assertContainerEqual(read_pe.pose_estimation_series["node3"], pose_estimation_series[2])
            self.assertContainerEqual(read_pe.skeleton, skeleton)
            self.assertEqual(len(read_pe.devices), 1)
            self.assertContainerEqual(read_pe.devices[0], self.nwbfile.devices["camera1"])


class TestPoseEstimationRoundtripPyNWB(NWBH5IOFlexMixin, TestCase):
    """Complex, more complete roundtrip test for PoseEstimation using pynwb.testing infrastructure."""

    def getContainerType(self):
        return "PoseEstimation"

    def addContainer(self):
        """Add the test PoseEstimation to the given NWBFile"""
        mock_PoseEstimation(nwbfile=self.nwbfile)

    def getContainer(self, nwbfile: NWBFile):
        return nwbfile.processing["behavior"]["PoseEstimation"]


class TestPoseTrainingRoundtripPyNWB(NWBH5IOFlexMixin, TestCase):
    """Complex, more complete roundtrip test for PoseTraining using pynwb.testing infrastructure."""

    def getContainerType(self):
        return "PoseTraining"

    def addContainer(self):
        """Add the test PoseTraining to the given NWBFile"""
        skeleton1 = mock_Skeleton(name="subject1")
        skeleton2 = mock_Skeleton(name="subject2")
        skeletons = Skeletons(skeletons=[skeleton1, skeleton2])

        source_video = mock_source_video(name="source_video")
        sk1_instance10 = mock_SkeletonInstance(id=np.uint(10), skeleton=skeleton1)
        sk1_instance11 = mock_SkeletonInstance(id=np.uint(11), skeleton=skeleton1)
        sk1_instances = mock_SkeletonInstances(skeleton_instances=[sk1_instance10, sk1_instance11])
        sk1_training_frame = mock_TrainingFrame(
            name="skeleton1_frame10",
            skeleton_instances=sk1_instances,
            source_video=source_video,
            source_video_frame_index=np.uint(10),
        )
        sk2_instance10 = mock_SkeletonInstance(id=np.uint(10), skeleton=skeleton2)
        sk2_instance11 = mock_SkeletonInstance(id=np.uint(11), skeleton=skeleton2)
        sk2_instance12 = mock_SkeletonInstance(id=np.uint(12), skeleton=skeleton2)
        sk2_instances = mock_SkeletonInstances(skeleton_instances=[sk2_instance10, sk2_instance11, sk2_instance12])
        sk2_training_frame = mock_TrainingFrame(
            name="skeleton2_frame10",
            skeleton_instances=sk2_instances,
            source_video=source_video,
            source_video_frame_index=np.uint(10),
        )

        training_frames = TrainingFrames(training_frames=[sk1_training_frame, sk2_training_frame])
        source_videos = SourceVideos(image_series=[source_video])

        pose_training = PoseTraining(
            training_frames=training_frames,
            source_videos=source_videos,
        )

        behavior_pm = self.nwbfile.create_processing_module(
            name="behavior",
            description="processed behavioral data",
        )
        behavior_pm.add(skeletons)
        behavior_pm.add(pose_training)

    def getContainer(self, nwbfile: NWBFile):
        return nwbfile.processing["behavior"]["PoseTraining"]



import warnings
from hdmf.utils import docval, popargs, get_docval, AllowPositional
from pynwb import register_class, TimeSeries, get_class
from pynwb.behavior import SpatialSeries
from pynwb.core import MultiContainerInterface

# TODO validate Skeleton nodes and edges correspondence, convert edges to uint
# TODO validate that all Skeleton nodes are used in edges
Skeleton = get_class("Skeleton", "ndx-pose")
Skeletons = get_class("Skeletons", "ndx-pose")
SkeletonInstance = get_class("SkeletonInstance", "ndx-pose")
SkeletonInstances = get_class("SkeletonInstances", "ndx-pose")
TrainingFrame = get_class("TrainingFrame", "ndx-pose")
TrainingFrames = get_class("TrainingFrames", "ndx-pose")
SourceVideos = get_class("SourceVideos", "ndx-pose")
PoseTraining = get_class("PoseTraining", "ndx-pose")


@register_class("PoseEstimationSeries", "ndx-pose")
class PoseEstimationSeries(SpatialSeries):
    """Estimated position (x, y) or (x, y, z) of a body part over time."""

    __nwbfields__ = ("confidence", "confidence_definition")

    # NOTE: custom mapper in ndx_pose.io.pose maps:
    # 'confidence' dataset -> 'definition' attribute in spec to 'confidence_definition' field in Python class
    # if not for the custom mapper, this class could be auto-generated from the spec

    @docval(
        {
            "name": "name",
            "type": str,
            "doc": "Name of this PoseEstimationSeries, usually the name of a body part.",
        },
        {
            "name": "data",
            "type": ("array_data", "data", TimeSeries),
            "shape": ((None, 2), (None, 3)),
            "doc": "Estimated position (x, y) or (x, y, z).",
        },
        {
            "name": "reference_frame",
            "type": str,
            "doc": "Description defining what the zero-position (0, 0) or (0, 0, 0) is.",
        },
        {
            "name": "confidence",
            "type": ("array_data", "data"),
            "shape": (None,),
            "doc": "Confidence or likelihood of the estimated positions, scaled to be between 0 and 1.",
            "default": None,
        },
        {
            "name": "unit",
            "type": str,
            "doc": (
                "Base unit of measurement for working with the data. The default value "
                "is 'pixels'. Actual stored values are not necessarily stored in these units. "
                "To access the data in these units, multiply 'data' by 'conversion'."
            ),
            "default": "pixels",
        },
        {
            "name": "confidence_definition",
            "type": str,
            "doc": "Description of how the confidence was computed, e.g., 'Softmax output of the deep neural network'.",
            "default": None,
        },
        *get_docval(
            TimeSeries.__init__,
            "conversion",
            "resolution",
            "offset",
            "timestamps",
            "starting_time",
            "rate",
            "comments",
            "description",
            "control",
            "control_description",
        ),
        allow_positional=AllowPositional.ERROR,
    )
    def __init__(self, **kwargs):
        """Construct a new PoseEstimationSeries representing pose estimates for a particular body part."""
        confidence, confidence_definition = popargs("confidence", "confidence_definition", kwargs)
        super().__init__(**kwargs)
        self.confidence = confidence
        self.confidence_definition = confidence_definition


@register_class("PoseEstimation", "ndx-pose")
# NOTE: NWB MultiContainerInterface extends NWBDataInterface and HDMF MultiContainerInterface
class PoseEstimation(MultiContainerInterface):
    """Estimated position data for multiple body parts, computed from the same video with the same tool/algorithm.
    The timestamps of each child PoseEstimationSeries type should be the same.
    """

    __clsconf__ = [
        {
            "add": "add_pose_estimation_series",
            "get": "get_pose_estimation_series",
            "create": "create_pose_estimation_series",
            "type": PoseEstimationSeries,
            "attr": "pose_estimation_series",
        },
        # NOTE: devices is a list of **linked** Device objects. Because they are linked, we do not set up
        # MultiContainerInterface-generated functions for devices.
    ]

    __nwbfields__ = (
        "description",
        "original_videos",
        "labeled_videos",
        "dimensions",
        "devices",
        "scorer",
        "source_software",
        "source_software_version",
        "nodes",
        "edges",
        "skeleton",  # <-- this is a link to a Skeleton object
    )

    # custom mapper in ndx_pose.io.pose maps:
    # 'source_software' dataset -> 'version' attribute to 'source_software_version' field
    # if not for the custom mapper and custom validation, this class could be auto-generated from the spec

    @docval(  # all fields are optional
        {
            "name": "pose_estimation_series",
            "type": ("array_data", "data"),
            "doc": "Estimated position data for each body part.",
            "default": None,
        },
        {
            "name": "name",
            "type": str,
            "doc": "Description of the pose estimation procedure and output.",
            "default": "PoseEstimation",
        },
        {
            "name": "description",
            "type": str,
            "doc": "Description of the pose estimation procedure and output.",
            "default": None,
        },
        {
            "name": "original_videos",
            "type": ("array_data", "data"),
            "shape": (None,),
            "doc": "Paths to the original video files. The number of files should equal the number of camera devices.",
            "default": None,
        },
        {
            "name": "labeled_videos",
            "type": ("array_data", "data"),
            "shape": (None,),
            "doc": "Paths to the labeled video files. The number of files should equal the number of camera devices.",
            "default": None,
        },
        {
            "name": "dimensions",
            "type": ("array_data", "data"),
            "shape": ((None, 2)),
            "doc": (
                "Dimensions of each labeled video file. The number of dimension pairs should equal the number of "
                "camera devices."
            ),
            "default": None,
        },
        {
            "name": "devices",
            "type": ("array_data", "data"),
            "doc": "Cameras used to record the videos.",
            "default": None,
        },
        {
            "name": "scorer",
            "type": str,
            "doc": "Name of the scorer / algorithm used.",
            "default": None,
        },
        {
            "name": "source_software",
            "type": str,
            "doc": "Name of the software tool used. Specifying the version attribute is strongly encouraged.",
            "default": None,
        },
        {
            "name": "source_software_version",
            "type": str,
            "doc": "Version string of the software tool used.",
            "default": None,
        },
        {
            "name": "skeleton",
            "type": Skeleton,
            "doc": (
                "Layout of body part locations and connections. The Skeleton object should be placed in a "
                "Skeletons object which resides in the NWBFile at the same level as the PoseEstimation object. "
                "The Skeleton object should be linked here."
            ),
            "default": None,
        },
        {
            "name": "nodes",
            "type": ("array_data", "data"),
            "doc": (
                "DEPRECATED. Please use the 'skeleton' argument instead. "
                "Array of body part names corresponding to the names of the PoseEstimationSeries objects within "
                "this container."
            ),
            "default": None,
        },
        {
            "name": "edges",
            "type": ("array_data", "data"),
            "doc": (
                "DEPRECATED. Please use the 'skeleton' argument instead. "
                "Array of pairs of indices corresponding to edges between nodes. Index values correspond to row "
                "indices of the 'nodes' field. Index values use 0-indexing."
            ),
            "default": None,
        },
        allow_positional=AllowPositional.ERROR,
    )
    def __init__(self, **kwargs):
        nodes, edges, skeleton = popargs("nodes", "edges", "skeleton", kwargs)
        if nodes is not None or edges is not None:
            if skeleton is not None:
                raise ValueError("Cannot specify 'skeleton' with 'nodes' or 'edges'.")
            # TODO: this Skeleton is normally a link to a Skeleton elsewhere in the file (e.g., in a Skeletons object)
            # Here, the Skeleton is constructed from the nodes and edges and exists only in this PoseEstimation object
            # and not as a child; this can have unintended consequences if the file is rewritten with the latest
            # schema. This is a limitation of the current implementation, and will be addressed in a future release.
            skeleton = Skeleton(name="subject", nodes=nodes, edges=edges)
            # warn on new, no warning on construction from existing file
            if not self._in_construct_mode:
                msg = (
                    "The 'nodes' and 'edges' constructor arguments are deprecated. Please use the 'skeleton' "
                    "argument instead. These will be removed in a future release."
                )
                warnings.warn(msg, DeprecationWarning)

        # devices must be added to the NWBFile before being linked to from a PoseEstimation object.
        # otherwise, they will be added as children of the PoseEstimation object.
        devices = popargs("devices", kwargs)
        if devices is not None:
            for device in devices:
                if device.parent is None:
                    raise ValueError(
                        "All devices linked to from a PoseEstimation object must be added to the NWBFile first."
                    )

        # validate that if original videos, labeled videos, or dimensions are provided, then an equal number of
        # camera devices must be provided.
        # specification of cameras was not allowed in ndx-pose 0.1.*.
        # warn on new, no warning on construction from existing file
        # TODO in a future release, change DeprecationWarning to a ValueError
        original_videos, labeled_videos, dimensions = popargs("original_videos", "labeled_videos", "dimensions", kwargs)
        if original_videos is not None and (devices is None or len(original_videos) != len(devices)):
            if not self._in_construct_mode:
                msg = (
                    "The number of original videos must equal the number of camera devices. This will become an error "
                    "in a future release."
                )
                raise DeprecationWarning(msg)
        if labeled_videos is not None and (devices is None or len(labeled_videos) != len(devices)):
            if not self._in_construct_mode:
                msg = (
                    "The number of labeled videos must equal the number of camera devices. This will become an error "
                    "in a future release."
                )
                raise DeprecationWarning(msg)
        if dimensions is not None and (devices is None or len(dimensions) != len(devices)):
            if not self._in_construct_mode:
                msg = (
                    "The number of dimensions must equal the number of camera devices. This will become an error in a "
                    "future release."
                )
                raise DeprecationWarning(msg)

        pose_estimation_series, description = popargs("pose_estimation_series", "description", kwargs)
        scorer = popargs("scorer", kwargs)
        source_software, source_software_version = popargs("source_software", "source_software_version", kwargs)
        super().__init__(**kwargs)

        self.pose_estimation_series = pose_estimation_series
        self.description = description
        self.original_videos = original_videos
        self.labeled_videos = labeled_videos
        self.dimensions = dimensions
        self.devices = devices
        self.scorer = scorer
        self.source_software = source_software
        self.source_software_version = source_software_version
        self.skeleton = skeleton

        # TODO include calibration images for 3D estimates?
        # TODO validate that the nodes correspond to the names of the pose estimation series objects

    @property
    def nodes(self):
        return self.skeleton.nodes

    @nodes.setter
    def nodes(self, value):
        raise ValueError(
            "Setting PoseEstimation.nodes is deprecated. Please use PoseEstimation.skeleton.nodes instead."
        )

    @property
    def edges(self):
        return self.skeleton.edges

    @edges.setter
    def edges(self, value):
        raise ValueError(
            "Setting PoseEstimation.edges is deprecated. Please use PoseEstimation.skeleton.edges instead."
        )



"""Test reading and validating NWB files generated with previous versions of this package."""

import numpy as np
import numpy.testing as npt
from pathlib import Path
from pynwb import NWBHDF5IO, validate
import pytest
import warnings

# NOTE: if this package is not imported, then the custom containers and mappers will not be used
import ndx_pose  # noqa: F401


def get_io(path):
    """Get an NWBHDF5IO object for the given path."""
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=r"Ignoring cached namespace .*",
            category=UserWarning,
        )
        return NWBHDF5IO(str(path), "r")


def test_0_1_1_poseestimation_nodes_edges():
    """Test that PoseEstimation objects with nodes and edges written before 0.2.0 are read correctly."""
    f = Path(__file__).parent / "0.1.1_poseestimation_nodes_edges.nwb"
    with get_io(f) as io:
        read_nwbfile = io.read()
        npt.assert_array_equal(
            read_nwbfile.processing["behavior"]["PoseEstimation"].skeleton.nodes[:],
            ["front_left_paw", "front_right_paw"],
        )
        npt.assert_array_equal(
            read_nwbfile.processing["behavior"]["PoseEstimation"].skeleton.edges[:], np.array([[0, 1]], dtype="uint8")
        )


@pytest.mark.parametrize(
    "file_path,expected_warnings,expected_errors",
    [
        (
            Path(__file__).parent / "0.1.1_poseestimation_nodes_edges.nwb",
            [],
            [],
        ),
        (
            Path(__file__).parent / "0.1.1_poseestimation_no_cameras.nwb",
            [],
            [],
        ),
    ],
)
def test_read(file_path, expected_warnings, expected_errors):
    """Test reading and validating NWB files generated with previous versions of this package."""
    with warnings.catch_warnings(record=True) as warnings_on_read:
        warnings.simplefilter("always")
        with get_io(file_path) as io:
            errors = validate(io)
            io.read()
            # NOTE: this does not error if the expected warnings are not present
            for w in warnings_on_read:
                if str(w.message) not in expected_warnings:
                    raise Exception("Unexpected warning: %s: %s" % (file_path, str(w.message)))
            if errors:
                unexpected_errors = []
                for e in errors:
                    if str(e) not in expected_errors:
                        warnings.warn("%s: %s" % (file_path, e))
                        unexpected_errors.append(e)
                if unexpected_errors:
                    raise Exception("%d validation error(s). See warnings." % len(unexpected_errors))



"""
Example script that demonstrates how to write a PoseEstimation and Skeletons object to an NWB file.


With one camera, one video, one skeleton, and three body parts per skeleton.
"""

import datetime
import numpy as np
from pynwb import NWBFile, NWBHDF5IO
from pynwb.file import Subject
from ndx_pose import (
    PoseEstimationSeries,
    PoseEstimation,
    Skeleton,
    Skeletons,
)

# initialize an NWBFile object
nwbfile = NWBFile(
    session_description="session_description",
    identifier="identifier",
    session_start_time=datetime.datetime.now(datetime.timezone.utc),
)

# add a subject to the NWB file
subject = Subject(subject_id="subject1", species="Mus musculus")
nwbfile.subject = subject

# create a skeleton that define the relationship between the markers. also link this skeleton to the subject.
skeleton = Skeleton(
    name="subject1_skeleton",
    nodes=["front_left_paw", "body", "front_right_paw"],
    # define edges between nodes using the indices of the nodes in the node list.
    # this array represents an edge between front left paw and body, and an edge between body and front right paw.
    edges=np.array([[0, 1], [1, 2]], dtype="uint8"),
    subject=subject,
)

# store the skeleton into a Skeletons container object.
# (this is more useful if you have multiple skeletons in your training data)
skeletons = Skeletons(skeletons=[skeleton])

# create a device for the camera
camera1 = nwbfile.create_device(
    name="camera1",
    description="camera for recording behavior",
    manufacturer="my manufacturer",
)

# a PoseEstimationSeries represents the estimated position of a single marker.
# in this example, we have three PoseEstimationSeries: one for the body and one for each front paw.
data = np.random.rand(100, 2)  # num_frames x (x, y) but can be (x, y, z)
timestamps = np.linspace(0, 10, num=100)  # a timestamp for every frame
confidence = np.random.rand(100)  # a confidence value for every frame
reference_frame = "(0,0,0) corresponds to ..."
confidence_definition = "Softmax output of the deep neural network."

front_left_paw = PoseEstimationSeries(
    name="front_left_paw",
    description="Marker placed around fingers of front left paw.",
    data=data,
    unit="pixels",
    reference_frame=reference_frame,
    timestamps=timestamps,
    confidence=confidence,
    confidence_definition=confidence_definition,
)

data = np.random.rand(100, 2)  # num_frames x (x, y) but can be (x, y, z)
confidence = np.random.rand(100)  # a confidence value for every frame
body = PoseEstimationSeries(
    name="body",
    description="Marker placed on center of body.",
    data=data,
    unit="pixels",
    reference_frame=reference_frame,
    timestamps=front_left_paw,  # link to timestamps of front_left_paw so we don't have to duplicate them
    confidence=confidence,
    confidence_definition=confidence_definition,
)

data = np.random.rand(100, 2)  # num_frames x (x, y) but can be num_frames x (x, y, z)
confidence = np.random.rand(100)  # a confidence value for every frame
front_right_paw = PoseEstimationSeries(
    name="front_right_paw",
    description="Marker placed around fingers of front right paw.",
    data=data,
    unit="pixels",
    reference_frame=reference_frame,
    timestamps=front_left_paw,  # link to timestamps of front_left_paw so we don't have to duplicate them
    confidence=confidence,
    confidence_definition=confidence_definition,
)

# store all PoseEstimationSeries in a list
pose_estimation_series = [front_left_paw, body, front_right_paw]

# create a PoseEstimation object that represents the estimated positions of each node, references
# the original video and labeled video files, and provides metadata on how these estimates were generated.
# multiple videos and cameras can be referenced.
pose_estimation = PoseEstimation(
    name="PoseEstimation",
    pose_estimation_series=pose_estimation_series,
    description="Estimated positions of front paws of subject1 using DeepLabCut.",
    original_videos=["path/to/camera1.mp4"],
    labeled_videos=["path/to/camera1_labeled.mp4"],
    dimensions=np.array([[640, 480]], dtype="uint16"),  # pixel dimensions of the video
    devices=[camera1],
    scorer="DLC_resnet50_openfieldOct30shuffle1_1600",
    source_software="DeepLabCut",
    source_software_version="2.3.8",
    skeleton=skeleton,  # link to the skeleton object
)

# create a "behavior" processing module to store the PoseEstimation and Skeletons objects
behavior_pm = nwbfile.create_processing_module(
    name="behavior",
    description="processed behavioral data",
)
behavior_pm.add(skeletons)
behavior_pm.add(pose_estimation)

# write the NWBFile to disk
path = "test_pose.nwb"
with NWBHDF5IO(path, mode="w") as io:
    io.write(nwbfile)

# read the NWBFile from disk and print out the PoseEstimation and Skeleton objects
# as well as the first training frame
with NWBHDF5IO(path, mode="r") as io:
    read_nwbfile = io.read()
    print(read_nwbfile.processing["behavior"]["PoseEstimation"])
    print(read_nwbfile.processing["behavior"]["Skeletons"]["subject1_skeleton"])



import os
from pynwb import load_namespaces

try:
    from importlib.resources import files
except ImportError:
    # TODO: Remove when python 3.9 becomes the new minimum
    from importlib_resources import files

# Get path to the namespace.yaml file with the expected location when installed not in editable mode
__location_of_this_file = files(__name__)
__spec_path = __location_of_this_file / "spec" / "ndx-pose.namespace.yaml"

# If that path does not exist, we are likely running in editable mode. Use the local path instead
if not os.path.exists(__spec_path):
    __spec_path = __location_of_this_file.parent.parent.parent / "spec" / "ndx-pose.namespace.yaml"

# Load the namespace
load_namespaces(str(__spec_path))

from . import io as __io  # import to register custom I/O mappers
from .pose import (
    PoseEstimation,
    PoseEstimationSeries,
    Skeleton,
    Skeletons,
    TrainingFrame,
    TrainingFrames,
    SkeletonInstance,
    SkeletonInstances,
    SourceVideos,
    PoseTraining,
)

__all__ = [
    "PoseEstimation",
    "PoseEstimationSeries",
    "Skeleton",
    "Skeletons",
    "TrainingFrame",
    "TrainingFrames",
    "SkeletonInstance",
    "SkeletonInstances",
    "SourceVideos",
    "PoseTraining",
]

# Remove these functions from the package
del load_namespaces
